'******************************************************************************
' C Function Library in Spin
' Copyright (c) 2010 - 2014 Dave Hein
' See end of file for terms of use.
'******************************************************************************
'******************************************************************************
' Revison History
' v1.0 - 4/2/2010 First official release
'******************************************************************************
{{ 
  This object provides some standard C library functions written in Spin.
  There are three main types of functions provided -- string manipulation,
  formatted I/O and memory allocation.  C functions that use a variable
  number of arguments are supported in two ways.  The first way uses
  unique function names for each number of arguments that is used, such
  as printf3 for a printf with three arguments.  The second way uses an
  argument list, such as vprintf.  The argument list consists of an array
  of longs that holds the values of the arguments.

  The start function must be called before using any of the serial I/O or
  memory allocation functions.  It will setup a a serial port that transmits on
  pin 30 and receives on pin 31 at 57600 baud.  It will also establish a stack
  space of 200 longs for the top object.  The malloc heap begins immediately
  after the stack space, and extends to the end of the RAM.

  Serial I/O and stack space parameters can be specified by calling start1
  instead of start.  The first four parameters of start1 are the same as those
  used in a call to the FullDuplexSerial start function.  The fifth parameter,
  stacksize, defines the size in longs of the stack space.

  Additional serial ports can be created by calling the openserial routine.
  In additional to rxpin, txpin, mode and baudrate the receive and transmit
  buffer sizes are also specified.  A fileinfo pointer is returned, which
  is used when calling fputc, fgetc and all of the other f... routines.

  The functions puts, putchar and getchar use the str1, tx1 and rx1 methods
  of a modified FullDuplexSerial object.  This object, cserial, allows for multiple
  serial ports through the use of a structer pointer called a handle.  See
  cserial.spin for more information.

  clib uses a file info structure to access I/O devices, such as the serial port.
  The I/O routines that start with the letter "f" require a pointer to the file info
  struction, wich is normally named stream.  Currently, only serial I/O and
  string I/O are supported.  Future releases will also support file I/O.
}}

OBJ
  mem : "cmalloc"
  fileio : "fsrwx"

CON
  NEW_LINE = 10
  SEEK_SET = 0
  SEEK_CUR = 1
  SEEK_END = 2

  stream_fileio = 2
  
  stream_type   = 0
  stream_data   = 1
  stream_size   = 8

  error_no_such_file        = 1
  error_not_enough_memory   = 2
  error_file_exists         = 3
  error_invalid_arg         = 4
  error_io_error            = 5
  error_not_directory       = 6
  error_is_directory        = 7
  error_directory_not_empty = 8
  error_name_too_long       = 9
  error_permission_denied   = 10

DAT
  currentwd long 0[25]
  tempwd long 0[25]
  errno long 0

'******************************************************************************  
' String Routines
'******************************************************************************  
PRI strcpy(dst, src) | slen
{{
  This routine copies the string pointed to by "src" to the location pointed
  to by "dst".  It returns the value of "dst".
}}
  slen := strsize(src) + 1
  bytemove(dst, src, slen)
  return dst

PRI strncpy(dst, src, num) | slen
{{
  This routine copies the first num bytes from the src string to the dst string.
  If src contains less than num bytes, then only the bytes contained in the src
  string are copied, and the remaining bytes are set to zero.  Note that if num
  is less than or equal to the string length of src the dst string will not be
  terminated with a NULL.  This routine returns the value of the dst pointer.
}}
  if (num < 1)
    return dst
  slen := strsize(src)
  if (slen > num)
    slen := num
  bytemove(dst, src, slen)
  bytefill(dst + slen, 0, num - slen)
  return dst

PUB fgetc(stream) | char, handle, num
{{
  This routine returns a single character from the input device defined by "stream".
  It will not return until a character has been received.
}}
  char := 0
  handle := long[stream][stream_data]
  num := fileio.hread(handle, @char, 1)
  if (num < 1)
    return -1
  return char

PUB fgets(str, num, stream) | handle, i, value
{{
  This routine returns a string from the input device defined by "stream".  It will
  return until it has either received a carriage return or "size" number of characters.
}}
  handle := long[stream][stream_data]
  result := str
  repeat i from 0 to num - 2
    value := fileio.hread(handle, str, 1)
    IF (value < 1)
      IF (i == 0)
        RETURN 0
      QUIT
    IF (BYTE[str] == 0)
      QUIT
    if byte[str++] == NEW_LINE
      quit
  BYTE[str] := 0

PRI ProcessPath(fname, pathstr, filestr) | fname1
  strncpy(pathstr, fname, 80)
  fname1 := SplitPathFile(pathstr)
  if byte[fname1] == 0 AND byte[pathstr] == "/"
    fname1 := filestr
    strcpy(fname1, pathstr + 1)
    byte[pathstr][1] := 0

  if fname1 <> pathstr
    strcpy(@tempwd, @currentwd)
    chdir(pathstr)

  return fname1

PUB fopen(fname, mode) | stream, handle, handlesize, value, pathstr[20], fname1, filestr[8]
{{
  This routine opens a file with the name pointed to by "fname".  The mode is
  determined by the first character of the string pointed to by "mode".  The first
  character must be "r", "w" or "a" for read, write or append.
  It returns a pointer to the file info structure if successful, or a NULL value
  if not successful.
}}
  fname1 := ProcessPath(fname, @pathstr, @filestr)

  ' Change directory if directory path is specified
  handlesize := fileio.handlesize

  ifnot (stream := mem.malloc(stream_size + handlesize))
    errno := error_not_enough_memory
    ChangeTempWD
    return 0
  handle := stream + stream_size
  long[stream][stream_data] := handle

  long[stream][stream_type] := stream_fileio
  value := fileio.hopen(handle, fname1, byte[mode])

  ChangeTempWD
  if value
    errno := -value
    mem.free(stream)
    return 0
  return stream

PRI ChangeTempWD
  if byte[@tempwd]
    chdir(@tempwd)
    byte[@tempwd] := 0

PRI SplitPathFile(pathstr) | fname, len, len1
  ' Search backwards for a "/"
  len1 := len := strsize(pathstr)
  ifnot len
    return pathstr
  if len == 1 and byte[pathstr] == "/"
    return pathstr + 1
  repeat while --len > 0
    if byte[pathstr][len] == "/"
      byte[pathstr][len] := 0
      return pathstr + len + 1
  if byte[pathstr] == "/"
    return pathstr + len1
  else
    return pathstr

PUB fread(buffer, elem_size, num, stream) | handle
{{
  This routine reads a block of data from the input device defined by "stream".
  It returns the number of bytes read, or a -1 if the end of file is
  encountered.
}}
  handle := long[stream][stream_data]
  result := fileio.hread(handle, buffer, elem_size * num)
  if result < 0
    result := 0
  
PUB fwrite(buffer, elem_size, num, stream) | handle
{{
  This routine writes a block of data to the output device defined by "stream".
  It returns the number of bytes written, or a -1 if there is an error.
}}
  handle := long[stream][stream_data]
  result := fileio.hwrite(handle, buffer, elem_size * num)

PUB fclose(stream) | handle
  handle := long[stream][stream_data]
  result := fileio.hclose(handle)
  ifnot result
    result := 1
  mem.free(stream)

PUB remove(fname)
  result := fileio.hopen(0, fname, "d")
  errno := -result
    
PUB rename(fname1, fname2)
  result := fileio.rename(fname1, fname2)
  errno := -result   

PUB create(fname) | handle
  handle := mem.malloc(fileio.handlesize)
  ifnot handle
    errno := error_not_enough_memory
    return -1
  ifnot (result := fileio.hopen(handle, fname, "r"))
    fileio.hclose(handle)
    result := 1
  elseif result == -1
    ifnot (result := fileio.hopen(handle, fname, "w"))
      fileio.hclose(handle)
      result := 1
  mem.free(handle)

PUB getfilesize(stream) | handle
  handle := long[stream][stream_data]
  result := fileio.hget_filesize(handle)

PUB mount_explicit(DO, CLK, DI, CS)
  result := fileio.mount_explicit(DO, CLK, DI, CS)
  bytemove(@currentwd, string("/"), 2)
  ifnot result
    result := 1

PUB mount(basepin)
  return mount_explicit(basepin, basepin + 1, basepin + 2, basepin + 3)

PUB getcwd(str, num)
  strncpy(str, @currentwd, num)
  return str

PUB FindChar(str, value_0)
  REPEAT WHILE (BYTE[str])
    IF (BYTE[str] == value_0)
      QUIT
    str++
  RETURN str

PUB chdir(path) | abspath[25], cwdlen, abslen, fname
  ' Check if current directory
  if byte[path] == 0 OR strcomp(path, string("."))
    return 0
  if strcomp(path, string("./"))
    return 0

  ' Check for absolute path
  if byte[path] == "/"
    strcpy(@abspath, path)
    abslen := strsize(@abspath)
    cwdlen := strsize(@currentwd)
    path := @abspath + cwdlen
    ' Check if relative from CWD
    if abslen > cwdlen AND byte[path] == "/"
      byte[path] := 0
      if strcomp(@currentwd, @abspath)
        byte[path++] := "/"
      else
        byte[path] := "/"
        path := @abspath
    else
      path := @abspath

  ' Check if going up one level
  elseif strcomp(path, string(".."))
    path := @abspath
    strcpy(path, @currentwd)
    fname := SplitPathFile(path)
    if byte[fname] == 0 AND byte[path] == "/"
      strcpy(path, string("/"))

  ' Else, relative path to current working directory
  else
    strcpy(@abspath, @currentwd)
    abslen := strsize(@abspath)
    if byte[@abspath + abslen - 1] <> "/"
      byte[@abspath + abslen++] := "/"
    strcpy(@abspath + abslen, path)
    path := @abspath + abslen

  ifnot (result := fileio.chdir(path))
    strcpy(@currentwd, @abspath)
    
  errno := -result

PUB opendir(path) | handlesize, stream, handle, value
  strcpy(@tempwd, @currentwd)
  if chdir(path)
    return 0
  handlesize := fileio.handlesize
  ifnot (stream := mem.malloc(stream_size + handlesize))
    errno := error_not_enough_memory
    ChangeTempWD
    return 0
  handle := stream + stream_size
  long[stream][stream_data] := handle
  long[stream][stream_type] := stream_fileio
  value := fileio.hopendir(handle)
  ChangeTempWD
  if value
    mem.free(stream)
    return 0
  return stream

PUB readdir(stream) | handle
  handle := long[stream][stream_data]
  result := fileio.hreaddir(handle)

PUB closedir(stream) | handle
  handle := long[stream][stream_data]
  result := fileio.hclosedir(handle)
  mem.free(stream)

PUB ftell(stream) | handle
  handle := long[stream][stream_data]
  result := fileio.htell(handle)

PUB fseek(stream, position, origin) | handle
{{
  This routine moves to the location in the file determined by
  the "newpos" and the "origin".  For origin values of 0, 1 or 2
  the base position is the start of the file, the current position
  or the end of the file respectively.
}}
  handle := long[stream][stream_data]
  if origin == SEEK_CUR
    position += fileio.htell(handle)
  elseif origin == SEEK_END
    position += fileio.hget_filesize(handle)
  result := fileio.hseek(handle, position)

PUB fflush(stream) | handle
  handle := long[stream][stream_data]
  result := fileio.hflush(handle)

PUB mkdir(fname) | stream, dummy
  result := fileio.mkdir(fname)
  errno := -result

PUB chmod(fname, modebits) | fname1, pathstr[20], filestr[8]
  fname1 := ProcessPath(fname, @pathstr, @filestr)
  result := fileio.chmod(fname1, modebits)
  ChangeTempWD

PUB getmod(fname) | fname1, pathstr[20], filestr[8]
  fname1 := ProcessPath(fname, @pathstr, @filestr)
  result := fileio.getmod(fname1)
  ChangeTempWD

PUB geterrno
  result := errno

{
+--------------------------------------------------------------------
|  TERMS OF USE: MIT License
+--------------------------------------------------------------------
Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files
(the "Software"), to deal in the Software without restriction,
including without limitation the rights to use, copy, modify, merge,
publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so,
subject to the following conditions:

The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
+------------------------------------------------------------------
}